from unittest import TestCase
from hamcrest import *

from cis.data_io.products.NCAR_NetCDF_RAF import NCAR_NetCDF_RAF_variable_name_selector
from cis.exceptions import InvalidVariableError


class TestDataReader(TestCase):
    class MockVar(object):
        def __init__(self, dimensions=('Time',)):
            self.dimensions = dimensions

    def test_GIVEN_no_attributes_or_variables_WHEN_construct_THEN_throw_exception(self):
        variables = []
        attributes = {}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No variables in the file so the type of data is unknown"))

    def test_GIVEN_no_attributes_WHEN_construct_THEN_throw_exception(self):
        variables = [{"Var": self.MockVar()}]
        attributes = {}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No attributes in the file so type of data is unknown"))

    def test_GIVEN_no_time_coordinate_or_general_coordinates_WHEN_construct_THEN_throw_exception(self):
        variables = [{"Var": self.MockVar()}]
        attributes = {"not time": "not time"}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No attributes indicating time variable name, "
                                              "expecting either 'Time_Coordinate' or 'Coordinates'"))

    def test_GIVEN_time_coordinate_but_no_variable_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        variables = [{"not time var": self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("There is no variable for the time co-ordinate '{}'"
                                              .format(expected_time_var)))

    def test_GIVEN_time_coordinate_but_no_lat_variable_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        variables = [{expected_time_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No attributes indicating latitude, "
                                              "expecting 'Station_Lat' or 'Longitude_Coordinate'"))

    def test_GIVEN_time_coordinate_station_lat_but_no_lon_variable_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        variables = [{expected_time_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Station_Lat": 27.2}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No attributes indicating longitude, expecting 'Station_Lon'"))

    def test_GIVEN_time_coordinate_station_lat_and_lon_variable_WHEN_construct_THEN_names_are_correct(self):
        expected_time_var = "time_var"
        expected_station_latitude = 27.1
        expected_station_longitude = 10
        variables = [{expected_time_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Station_Lat": expected_station_latitude,
                      "Station_Lon": expected_station_longitude}
        aggregated_dim = "Time"
        expected_time_dimensions = (aggregated_dim,)

        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(decider.station, is_(True), "File should be for station")
        assert_that(decider.time_variable_name, is_(expected_time_var))
        assert_that(decider.station_latitude, is_([expected_station_latitude]))
        assert_that(decider.station_longitude, is_([expected_station_longitude]))
        assert_that(decider.altitude, is_([0]))
        assert_that(decider.time_dimensions, is_(expected_time_dimensions))

    def test_GIVEN_station_alt_WHEN_construct_THEN_alt_set(self):
        expected_time_var = "time_var"
        expected_station_latitude = 27.1
        expected_station_longitude = 10
        expected_alt = 125.4
        variables = [{expected_time_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Station_Lat": expected_station_latitude,
                      "Station_Lon": expected_station_longitude,
                      "Station_Altitude": expected_alt}

        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(decider.altitude, is_([expected_alt]))

    def test_GIVEN_station_coords_in_different_cases_WHEN_construct_THEN_values_are_correct(
            self):
        expected_time_var = "time_var"
        expected_station_latitude = 27.1
        expected_station_longitude = 10
        expected_alt = 125.4
        variables = [{expected_time_var: self.MockVar()}]
        attributes = {"tIME_CoorDinate": expected_time_var,
                      "StatioN_lat": expected_station_latitude,
                      "StaTion_Lon": expected_station_longitude,
                      "StAtion_Altitude": expected_alt}

        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(decider.station, is_(True), "File should be for station")
        assert_that(decider.time_variable_name, is_(expected_time_var))
        assert_that(decider.station_latitude, is_([expected_station_latitude]))
        assert_that(decider.station_longitude, is_([expected_station_longitude]))
        assert_that(decider.altitude, is_([expected_alt]))

    def test_GIVEN_time_coordinate_variable_lat_but_lat_variable_does_not_exist_WHEN_construct_THEN_throw_exception(
            self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        variables = [{expected_time_var: self.MockVar(), "not LON": self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("There is no variable for the latitude co-ordinate '{}'"
                                              .format(expected_lat_var)))

    def test_GIVEN_time_coordinate_variable_lat_but_no_lon_variable_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("No attributes indicating longitude variable name, "
                                              "expecting 'Longitude_Coordinate'"))

    def test_GIVEN_time_coordinate_variable_lat_lon_but_lon_variable_does_not_exist_WHEN_construct_THEN_throw_exception(
            self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(), "not LAT": self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("There is no variable for the longitude co-ordinate '{}'"
                                              .format(expected_lon_var)))

    def test_GIVEN_time_coordinate_variable_lat_lon_with_no_alt_WHEN_construct_THEN_values_are_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.station, is_(False), "station")
        assert_that(selector.time_variable_name, is_(expected_time_var), "time variable name")
        assert_that(selector.longitude_variable_name, is_(expected_lon_var), "longitude variable name")
        assert_that(selector.latitude_variable_name, is_(expected_lat_var), "latitude variable name")
        assert_that(selector.altitude, is_([0]), "altitude is fixed at zero")
        assert_that(selector.pressure_variable_name, is_(None), "pressure variable name")

    def test_GIVEN_time_coordinate_var_lat_lon_alt_but_alt_variable_does_not_exist_WHEN_construct_THEN_throw_exception(
            self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      "not ALT": self.MockVar(), expected_lon_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var,
                      "Vertical_Coordinate": expected_alt_var}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0],
                    is_("There is no variable for the vertical co-ordinate '{}'".format(expected_alt_var)))

    def test_GIVEN_time_coordinate_variable_lat_lon_alt_WHEN_construct_THEN_values_are_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_alt_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var,
                      "Vertical_Coordinate": expected_alt_var}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.altitude, is_(None), "altitude not fixed")
        assert_that(selector.altitude_variable_name, is_(expected_alt_var), "altitude var name")

    def test_GIVEN_time_coordinate_variable_lat_lon_with_PSXC_WHEN_construct_THEN_pressure_name_is_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_pressure_var = "PSXC"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_pressure_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.pressure_variable_name, is_(expected_pressure_var), "pressure variable name")

    def test_GIVEN_time_coordinate_variable_lat_lon_with_PRE_WHEN_construct_THEN_pressure_name_is_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_pressure_var = "PRE"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_pressure_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.pressure_variable_name, is_(expected_pressure_var), "pressure variable name")

    def test_GIVEN_time_coordinate_variable_lat_lon_with_PRE_and_PSXC_WHEN_construct_THEN_pressure_name_is_set_to_PSXC(
            self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_pressure_var = "PSXC"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_pressure_var: self.MockVar(), "PRE": self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.pressure_variable_name, is_(expected_pressure_var), "pressure variable name")

    def test_GIVEN_coordinate_attribute_which_does_not_have_four_variables_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_alt_var: self.MockVar()}]
        attributes = {"Coordinates": "incorect"}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("The coordinate attribute does not have four entries. "
                                              "It should be space separated \"longitude latitude altitude time\""))

    def test_GIVEN_coordinate_attribute_but_no_variable_for_one_value_WHEN_construct_THEN_throw_exception(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"

        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), "NOT ALT": self.MockVar()}]
        attributes = {"Coordinates": "{} {} {} {}".format(expected_lon_var, expected_lat_var,
                                                          expected_alt_var, expected_time_var)}

        with self.assertRaises(InvalidVariableError) as cm:
            NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(cm.exception.args[0], is_("There is no variable for the co-ordinate '{}'".format(expected_alt_var)))

    def test_GIVEN_coordinate_attribute_WHEN_construct_THEN_values_are_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_alt_var: self.MockVar()}]
        attributes = {"Coordinates": "{} {} {} {}".format(expected_lon_var, expected_lat_var,
                                                          expected_alt_var, expected_time_var)}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.station, is_(False), "station")
        assert_that(selector.time_variable_name, is_(expected_time_var), "time variable name")
        assert_that(selector.longitude_variable_name, is_(expected_lon_var), "longitude variable name")
        assert_that(selector.latitude_variable_name, is_(expected_lat_var), "latitude variable name")
        assert_that(selector.altitude, is_(None), "altitude not fixed")
        assert_that(selector.altitude_variable_name, is_(expected_alt_var), "altitude var name")
        assert_that(selector.pressure_variable_name, is_(None), "pressure variable name")

    def test_GIVEN_coordinate_attribute_and_pressure_WHEN_construct_THEN_pressure_name_is_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        expected_pres_var = "PSXC"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_alt_var: self.MockVar(),
                      expected_pres_var: self.MockVar()}]
        attributes = {"Coordinates": "{} {} {} {}".format(expected_lon_var, expected_lat_var,
                                                          expected_alt_var, expected_time_var)}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.station, is_(False), "station")
        assert_that(selector.time_variable_name, is_(expected_time_var), "time variable name")
        assert_that(selector.longitude_variable_name, is_(expected_lon_var), "longitude variable name")
        assert_that(selector.latitude_variable_name, is_(expected_lat_var), "latitude variable name")
        assert_that(selector.altitude, is_(None), "altitude not fixed")
        assert_that(selector.altitude_variable_name, is_(expected_alt_var), "altitude var name")
        assert_that(selector.pressure_variable_name, is_(expected_pres_var), "pressure variable name")

    def test_GIVEN_time_coordinate_variable_lat_lon_alt_with_timestamp_WHEN_construct_THEN_timestamp_set(self):
        expected_time_var = "time_var"
        expected_lat_var = "LAT"
        expected_lon_var = "LON"
        expected_alt_var = "ALT"
        expected_timestamp = "timestamp"
        variables = [{expected_time_var: self.MockVar(), expected_lat_var: self.MockVar(),
                      expected_lon_var: self.MockVar(), expected_alt_var: self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Latitude_Coordinate": expected_lat_var,
                      "Longitude_Coordinate": expected_lon_var,
                      "Vertical_Coordinate": expected_alt_var,
                      "Time_Stamp_Info": expected_timestamp}

        selector = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(selector.time_stamp_info, is_([expected_timestamp]), "time stamp info")

    def test_GIVEN_valid_file_with_all_variables_same_shape_as_time_WHEN_get_vars_THEN_all_variables_returned(self):
        expected_time_var = "time_var"
        expected_station_latitude = "27.1"
        expected_station_longitude = "10"
        variables = [{expected_time_var: self.MockVar(), "another_var": self.MockVar()}]
        attributes = {"Time_Coordinate": expected_time_var,
                      "Station_Lat": expected_station_latitude,
                      "Station_Lon": expected_station_longitude}
        aggregated_dim = "Time"
        expected_time_dimensions = [aggregated_dim]
        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        vars = decider.get_variable_names_which_have_time_coord()

        assert_that(vars, all(variables), "File should be for station")

    def test_GIVEN_valid_file_with_one_variables_same_shape_others_not_WHEN_get_vars_THEN_only_same_shape_is_returned(
            self):
        time_dims = ["Time"]
        time_var = "time_var"
        expected_var = "good_shape"
        variable_dimensions = {
            time_var: time_dims,
            expected_var: time_dims,
            "too_many_dims": ["one", "two"],
            "too_few_dims": [],
            "Not time": ["not time"],
            "Time and one other": [time_dims[0], "other"]
        }
        attributes = {"Time_Coordinate": time_var,
                      "Station_Lat": "27.1",
                      "Station_Lon": "10"}

        variables = [{key: self.MockVar(val) for key, val in list(variable_dimensions.items())}]
        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        vars = decider.get_variable_names_which_have_time_coord()

        assert_that(vars, is_(set([expected_var, time_var, "Time and one other"])), "variables should be ones with same dim as time")

    def test_GIVEN_valid_file_w_1_var_same_shape_others_not_and_t_is_multid_WHEN_get_vars_THEN_only_same_shape_returned(
            self):
        time_dims = ["Time", "time2"]
        time_var = "time_var"
        expected_var = "good_shape"
        variable_dimensions = {
            time_var: time_dims,
            expected_var: time_dims,
            "too_many_dims": ["one", "two", "Three"],
            "too_few_dims": ["one"],
            "Not time": ["not time", "not Time"],
            "Time and one other": [time_dims[0], time_dims[1], "other"],
            "Time and Time": [time_dims[0], time_dims[0]],
            "Time2 and time": [time_dims[1], time_dims[0]]
        }
        attributes = {"Time_Coordinate": time_var,
                      "Station_Lat": "27.1",
                      "Station_Lon": "10"}

        variables = [{key: self.MockVar(val) for key, val in list(variable_dimensions.items())}]
        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        vars = decider.get_variable_names_which_have_time_coord()

        assert_that(vars, is_(set([expected_var, time_var, "Time and one other"])),
                    "variables should be ones with same dim as time")

    def test_GIVEN_valid_file_w_var_with_aux_coords_WHEN_find_auxiliary_coordinates_THEN_returns_aux_coords(self):
        time_dims = ["Time"]
        time_var = "time_var"
        expected_var = "good_shape"
        variable_dimensions = {
            time_var: time_dims,
            expected_var: time_dims,
            "extra_dim": [time_dims[0], "aux_coord"],
            "two_extra": [time_dims[0], "aux_coord", "another_aux"],
            "aux_coord": ["aux_coord"],
        }
        attributes = {"Time_Coordinate": time_var,
                      "Station_Lat": "27.1",
                      "Station_Lon": "10"}

        variables = [{key: self.MockVar(val) for key, val in list(variable_dimensions.items())}]
        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(decider.find_auxiliary_coordinate(expected_var), is_(None))
        assert_that(decider.find_auxiliary_coordinate('extra_dim'), is_('aux_coord'))

        # Don't return variables with more than one aux coord (which would have to be more than 2D)
        with self.assertRaises(InvalidVariableError) as cm:
            decider.find_auxiliary_coordinate('two_extra')

    def test_GIVEN_valid_file_w_var_with_unnamed_aux_WHEN_find_auxiliary_coordinates_THEN_returns_aux_coords(self):
        """
        Test if we can find the right variable even if the auxiliary dimension has a different name
        """
        time_dims = ["Time"]
        time_var = "time_var"
        expected_var = "good_shape"
        variable_dimensions = {
            time_var: time_dims,
            expected_var: time_dims,
            "extra_dim": [time_dims[0], "aux_coord"],
            "two_extra": [time_dims[0], "aux_coord", "another_aux"],
            "aux_var": ["aux_coord"]
        }
        attributes = {"Time_Coordinate": time_var,
                      "Station_Lat": "27.1",
                      "Station_Lon": "10"}

        variables = [{key: self.MockVar(val) for key, val in list(variable_dimensions.items())}]
        decider = NCAR_NetCDF_RAF_variable_name_selector(attributes, variables)

        assert_that(decider.find_auxiliary_coordinate(expected_var), is_(None))
        assert_that(decider.find_auxiliary_coordinate('extra_dim'), is_('aux_var'))

        # Don't return variables with more than one aux coord (which would have to be more than 2D)
        with self.assertRaises(InvalidVariableError) as cm:
            decider.find_auxiliary_coordinate('two_extra')



"""
Test Gridded - Gridded Collocation
"""
from unittest import TestCase

from nose.tools import istest, raises
from iris.exceptions import CoordinateNotFoundError
import numpy

from cis.exceptions import ClassNotFoundError
from cis.collocation.col_implementations import GriddedCollocator, gridded_gridded_nn, gridded_gridded_li, nn_p
import cis.data_io.gridded_data as gridded_data
from cis.test.util.mock import make_dummy_2d_cube, make_dummy_2d_cube_with_small_offset_in_lat_and_lon, \
    make_dummy_2d_cube_with_small_offset_in_lat, make_dummy_2d_cube_with_small_offset_in_lon, \
    make_list_with_2_dummy_2d_cubes_where_verticies_are_in_cell_centres, make_mock_cube


def does_coord_exist_in_cube(cube, coord):
    try:
        cube.coord(coord)
        return True
    except CoordinateNotFoundError:
        return False


class GriddedGriddedCollocatorTests(object):
    def setUp(self):
        self.collocator = None

    @istest
    @raises(ClassNotFoundError)
    def invalid_kernel_throws_error(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube())
        data_cube = gridded_data.make_from_cube(make_mock_cube())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=nn_p())[0]

    @istest
    def test_gridded_gridded_nn_for_same_grids_check_returns_original_data(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube())
        data_cube = gridded_data.make_from_cube(make_mock_cube())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lat_and_lon_using_nn(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(data_offset=100))
        data_cube = gridded_data.make_from_cube(make_mock_cube(horizontal_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert not numpy.array_equal(data_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert not numpy.array_equal(data_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lat_and_lon_different_grid_size_using_nn(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(data_offset=100))
        data_cube = gridded_data.make_from_cube(make_mock_cube(lat_dim_length=10,
                                                               lon_dim_length=6, horizontal_offset=0.0))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        result = numpy.array([[1, 3, 6],
                              [13, 15, 18],
                              [25, 27, 30],
                              [43, 45, 48],
                              [55, 57, 60]])

        assert numpy.array_equal(out_cube.data, result)

    @istest
    def test_gridded_gridded_for_two_grids_offset_by_half_grid_spacing_using_nn(self):
        sample_cube, data_cube = make_list_with_2_dummy_2d_cubes_where_verticies_are_in_cell_centres()
        sample_cube = gridded_data.make_from_cube(sample_cube)
        data_cube = gridded_data.make_from_cube(data_cube)

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        result = numpy.array([[0., 0., 1., 0., 1.],
                              [0., 0., 1., 0., 1.],
                              [1., 1., 0., 1., 0.],
                              [0., 0., 1., 0., 1.],
                              [1., 1., 0., 1., 0.]])

        assert numpy.array_equal(out_cube.data, result)

    @istest
    def test_gridded_gridded_li_for_same_grids_check_returns_original_data(self):
        sample_cube = gridded_data.make_from_cube(make_dummy_2d_cube())
        data_cube = gridded_data.make_from_cube(make_dummy_2d_cube())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert numpy.allclose(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lat_using_li(self):
        # Test fails on Iris 1.5.1., but passes on version 1.6.1
        sample_cube = gridded_data.make_from_cube(make_dummy_2d_cube())
        data_cube = gridded_data.make_from_cube(make_dummy_2d_cube_with_small_offset_in_lat())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert numpy.allclose(data_cube.data, out_cube.data, atol=0.1)
        assert not numpy.array_equal(sample_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lon_using_li(self):
        sample_cube = gridded_data.make_from_cube(make_dummy_2d_cube())
        data_cube = gridded_data.make_from_cube(make_dummy_2d_cube_with_small_offset_in_lon())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert numpy.allclose(data_cube.data, out_cube.data, atol=0.1)
        assert not numpy.array_equal(sample_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lat_and_lon_using_li(self):
        # Test fails on Iris 1.5.1., but passes on version 1.6.1
        sample_cube = gridded_data.make_from_cube(make_dummy_2d_cube())
        data_cube = gridded_data.make_from_cube(make_dummy_2d_cube_with_small_offset_in_lat_and_lon())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert numpy.allclose(data_cube.data, out_cube.data, atol=0.2)
        assert not numpy.array_equal(sample_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_for_one_grid_with_slight_offset_in_lat_and_lon_different_grid_size_using_li(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(data_offset=100))
        data_cube = gridded_data.make_from_cube(make_mock_cube(lat_dim_length=10,
                                                               lon_dim_length=6, horizontal_offset=0.0))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[1., 3.5, 6.],
                              [14.5, 17., 19.5],
                              [28., 30.5, 33.],
                              [41.5, 44., 46.5],
                              [55., 57.5, 60.]])

        assert numpy.allclose(out_cube.data, result)

    @istest
    def test_gridded_gridded_for_two_grids_offset_by_half_grid_spacing_using_li(self):
        # Test fails on Iris 1.5.1., but passes on version 1.6.1
        sample_cube, data_cube = make_list_with_2_dummy_2d_cubes_where_verticies_are_in_cell_centres()
        sample_cube = gridded_data.make_from_cube(sample_cube)
        data_cube = gridded_data.make_from_cube(data_cube)

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        expected_result = numpy.array([[-1.5, 0.5, 0.5, 0.5, 0.5],
                                       [0.5, 0.5, 0.5, 0.5, 0.5],
                                       [0.5, 0.5, 0.5, 0.5, 0.5],
                                       [0.5, 0.5, 0.5, 0.5, 0.5],
                                       [0.5, 0.5, 0.5, 0.5, 0.5]])

        assert numpy.array_equal(out_cube.data, expected_result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_nn_with_one_grid_containing_time(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube())
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(data_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_li_with_one_grid_containing_altitude(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube())
        data_cube = gridded_data.make_from_cube(make_mock_cube(alt_dim_length=7))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(data_cube.coord('altitude').points, out_cube.coord('altitude').points)

    @istest
    def test_gridded_gridded_nn_with_one_grid_containing_time_and_slightly_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, horizontal_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(data_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_one_grid_containing_time_and_moderate_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube())
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, horizontal_offset=2.6))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        expected_result = numpy.array([[[1., 2., 3., 4., 5., 6., 7.],
                                        [1., 2., 3., 4., 5., 6., 7.],
                                        [8., 9., 10., 11., 12., 13., 14.]],

                                       [[1., 2., 3., 4., 5., 6., 7.],
                                        [1., 2., 3., 4., 5., 6., 7.],
                                        [8., 9., 10., 11., 12., 13., 14.]],

                                       [[22., 23., 24., 25., 26., 27., 28.],
                                        [22., 23., 24., 25., 26., 27., 28.],
                                        [29., 30., 31., 32., 33., 34., 35.]],

                                       [[43., 44., 45., 46., 47., 48., 49.],
                                        [43., 44., 45., 46., 47., 48., 49.],
                                        [50., 51., 52., 53., 54., 55., 56.]],

                                       [[64., 65., 66., 67., 68., 69., 70.],
                                        [64., 65., 66., 67., 68., 69., 70.],
                                        [71., 72., 73., 74., 75., 76., 77.]]])

        assert numpy.array_equal(out_cube.data, expected_result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(data_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_li_with_one_grid_containing_time_and_slightly_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, horizontal_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[0.44, 1.44, 2.44, 3.44, 4.44, 5.44, 6.44],
                               [7.44, 8.44, 9.44, 10.44, 11.44, 12.44, 13.44],
                               [14.44, 15.44, 16.44, 17.44, 18.44, 19.44, 20.44]],

                              [[21.44, 22.44, 23.44, 24.44, 25.44, 26.44, 27.44],
                               [28.44, 29.44, 30.44, 31.44, 32.44, 33.44, 34.44],
                               [35.44, 36.44, 37.44, 38.44, 39.44, 40.44, 41.44]],

                              [[42.44, 43.44, 44.44, 45.44, 46.44, 47.44, 48.44],
                               [49.44, 50.44, 51.44, 52.44, 53.44, 54.44, 55.44],
                               [56.44, 57.44, 58.44, 59.44, 60.44, 61.44, 62.44]],

                              [[63.44, 64.44, 65.44, 66.44, 67.44, 68.44, 69.44],
                               [70.44, 71.44, 72.44, 73.44, 74.44, 75.44, 76.44],
                               [77.44, 78.44, 79.44, 80.44, 81.44, 82.44, 83.44]],

                              [[84.44, 85.44, 86.44, 87.44, 88.44, 89.44, 90.44],
                               [91.44, 92.44, 93.44, 94.44, 95.44, 96.44, 97.44],
                               [98.44, 99.44, 100.44, 101.44, 102.44, 103.44, 104.44]]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(data_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_sample_containing_time_and_pressure_and_small_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7,
                                                                 pres_dim_length=10, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(horizontal_offset=0.1, time_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert not does_coord_exist_in_cube(out_cube, 'time')
        assert not does_coord_exist_in_cube(out_cube, 'pressure')

    @istest
    def test_gridded_gridded_li_with_sample_grid_4d_and_slightly_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(alt_dim_length=7, time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(horizontal_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[0.92, 1.92, 2.92],
                              [3.92, 4.92, 5.92],
                              [6.92, 7.92, 8.92],
                              [9.92, 10.92, 11.92],
                              [12.92, 13.92, 14.92]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert not does_coord_exist_in_cube(out_cube, 'time')
        assert not does_coord_exist_in_cube(out_cube, 'altitude')


class TestGriddedGriddedCollocator(GriddedGriddedCollocatorTests, TestCase):
    """
    Contains tests that are not available to Iris Regrid collocator
    """

    def setUp(self):
        self.collocator = GriddedCollocator(extrapolate=True)

    @istest
    def test_gridded_gridded_nn_with_both_grids_containing_time_and_small_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7,
                                                               horizontal_offset=0.1, time_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_data_grid_containing_scalar_time(self):
        import iris.coords
        import datetime
        from cis.time_util import convert_datetime_to_std_time

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))
        data_cube = gridded_data.make_from_cube(make_mock_cube())

        data_cube.add_aux_coord(iris.coords.DimCoord(convert_datetime_to_std_time(datetime.datetime(1984, 8, 28)),
                                                     standard_name='time',
                                                     units='days since 1600-01-01 00:00:00'))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        time_units = out_cube.coord('time').units
        val = time_units.date2num(datetime.datetime(1984, 8, 28))

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal([val], out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_data_grid_containing_time_length_one(self):
        import numpy as np

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=1, time_offset=3))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        expected = np.array([[[1., 1., 1., 1., 1., 1., 1.], [2., 2., 2., 2., 2., 2., 2.], [3., 3., 3., 3., 3., 3., 3.]],
                             [[4., 4., 4., 4., 4., 4., 4.], [5., 5., 5., 5., 5., 5., 5.], [6., 6., 6., 6., 6., 6., 6.]],
                             [[7., 7., 7., 7., 7., 7., 7.], [8., 8., 8., 8., 8., 8., 8.], [9., 9., 9., 9., 9., 9., 9.]],
                             [[10., 10., 10., 10., 10., 10., 10.], [11., 11., 11., 11., 11., 11., 11.],
                              [12., 12., 12., 12., 12., 12., 12.]],
                             [[13., 13., 13., 13., 13., 13., 13.], [14., 14., 14., 14., 14., 14., 14.],
                              [15., 15., 15., 15., 15., 15., 15.]]])

        assert numpy.array_equal(expected, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_data_grid_containing_no_time(self):

        sample_cube = gridded_data.make_from_cube(make_mock_cube(horizontal_offset=1.0, time_dim_length=7))
        data_cube = gridded_data.make_from_cube(make_mock_cube())

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_nn_with_sample_grid_containing_no_time(self):

        sample_cube = gridded_data.make_from_cube(make_mock_cube(horizontal_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)

    @istest
    def test_gridded_gridded_li_with_data_grid_containing_time_length_one(self):
        import numpy as np

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=1, time_offset=1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        expected = np.array([[[1., 1., 1., 1., 1., 1., 1.], [2., 2., 2., 2., 2., 2., 2.], [3., 3., 3., 3., 3., 3., 3.]],
                             [[4., 4., 4., 4., 4., 4., 4.], [5., 5., 5., 5., 5., 5., 5.], [6., 6., 6., 6., 6., 6., 6.]],
                             [[7., 7., 7., 7., 7., 7., 7.], [8., 8., 8., 8., 8., 8., 8.], [9., 9., 9., 9., 9., 9., 9.]],
                             [[10., 10., 10., 10., 10., 10., 10.], [11., 11., 11., 11., 11., 11., 11.],
                              [12., 12., 12., 12., 12., 12., 12.]],
                             [[13., 13., 13., 13., 13., 13., 13.], [14., 14., 14., 14., 14., 14., 14.],
                              [15., 15., 15., 15., 15., 15., 15.]]])

        assert numpy.array_equal(expected, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)


    @istest
    def test_gridded_gridded_li_with_data_grid_containing_scalar_time(self):
        import iris.coords
        import datetime
        from cis.time_util import convert_datetime_to_std_time

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7))
        data_cube = gridded_data.make_from_cube(make_mock_cube())

        data_cube.add_aux_coord(iris.coords.DimCoord(convert_datetime_to_std_time(datetime.datetime(1984, 8, 28)),
                                                     standard_name='time',
                                                     units='days since 1600-01-01 00:00:00'))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        time_units = out_cube.coord('time').units
        val = time_units.date2num(datetime.datetime(1984, 8, 28))

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal([val], out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_both_grids_containing_time_and_moderate_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7,
                                                               horizontal_offset=2.6, time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        result = numpy.array([[[1., 1., 1., 2., 3., 4., 5.],
                               [1., 1., 1., 2., 3., 4., 5.],
                               [8., 8., 8., 9., 10., 11., 12.]],

                              [[1., 1., 1., 2., 3., 4., 5.],
                               [1., 1., 1., 2., 3., 4., 5.],
                               [8., 8., 8., 9., 10., 11., 12.]],

                              [[22., 22., 22., 23., 24., 25., 26.],
                               [22., 22., 22., 23., 24., 25., 26.],
                               [29., 29., 29., 30., 31., 32., 33.]],

                              [[43., 43., 43., 44., 45., 46., 47.],
                               [43., 43., 43., 44., 45., 46., 47.],
                               [50., 50., 50., 51., 52., 53., 54.]],

                              [[64., 64., 64., 65., 66., 67., 68.],
                               [64., 64., 64., 65., 66., 67., 68.],
                               [71., 71., 71., 72, 73., 74., 75.]]])

        assert numpy.array_equal(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_li_with_both_grids_containing_time_and_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7,
                                                               horizontal_offset=0.1, time_offset=0.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[-6.e-02, 9.4e-01, 1.94, 2.94, 3.94, 4.94, 5.94],
                               [6.94, 7.94, 8.94, 9.94, 1.094e+01, 1.194e+01, 1.294e+01],
                               [1.394e+01, 1.494e+01, 1.594e+01, 1.694e+01, 1.794e+01, 1.894e+01, 1.994e+01]],

                              [[2.094e+01, 2.194e+01, 2.294e+01, 2.394e+01, 2.494e+01, 2.594e+01, 2.694e+01],
                               [2.794e+01, 2.894e+01, 2.994e+01, 3.094e+01, 3.194e+01, 3.294e+01, 3.394e+01],
                               [3.494e+01, 3.594e+01, 3.694e+01, 3.794e+01, 3.894e+01, 3.994e+01, 4.094e+01]],

                              [[4.194e+01, 4.294e+01, 4.394e+01, 4.494e+01, 4.594e+01, 4.694e+01, 4.794e+01],
                               [4.894e+01, 4.994e+01, 5.094e+01, 5.194e+01, 5.294e+01, 5.394e+01, 5.494e+01],
                               [5.594e+01, 5.694e+01, 5.794e+01, 5.894e+01, 5.994e+01, 6.094e+01, 6.194e+01]],

                              [[6.294e+01, 6.394e+01, 6.494e+01, 6.594e+01, 6.694e+01, 6.794e+01, 6.894e+01],
                               [6.994e+01, 7.094e+01, 7.194e+01, 7.294e+01, 7.394e+01, 7.494e+01, 7.594e+01],
                               [7.694e+01, 7.794e+01, 7.894e+01, 7.994e+01, 8.094e+01, 8.194e+01, 8.294e+01]],

                              [[8.394e+01, 8.494e+01, 8.594e+01, 8.694e+01, 8.794e+01, 8.894e+01, 8.994e+01],
                               [9.094e+01, 9.194e+01, 9.294e+01, 9.394e+01, 9.494e+01, 9.594e+01, 9.694e+01],
                               [9.794e+01, 9.894e+01, 9.994e+01, 1.0094e+02, 1.0194e+02, 1.0294e+02, 1.0394e+02]]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_both_grids_containing_time_with_moderate_offset_and_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=10, lat_dim_length=6, time_dim_length=14, horizontal_offset=2.6,
            time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        result = numpy.array([[[1., 1., 1., 2., 3., 4., 5.],
                               [29., 29., 29., 30., 31., 32., 33.],
                               [99., 99., 99., 100., 101., 102., 103.]],

                              [[141., 141., 141., 142., 143., 144., 145.],
                               [169., 169., 169., 170., 171., 172., 173.],
                               [239., 239., 239., 240., 241., 242., 243.]],

                              [[281., 281., 281., 282., 283., 284., 285.],
                               [309., 309., 309., 310., 311., 312., 313.],
                               [379., 379., 379., 380., 381., 382., 383.]],

                              [[421., 421., 421., 422., 423., 424., 425.],
                               [449., 449., 449., 450., 451., 452., 453.],
                               [519., 519., 519., 520., 521., 522., 523.]],

                              [[561., 561., 561., 562., 563., 564., 565.],
                               [589., 589., 589., 590., 591., 592., 593.],
                               [659., 659., 659., 660., 661., 662., 663.]]])

        assert numpy.array_equal(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_li_with_both_grids_containing_time_with_moderate_offset_and_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=10, lat_dim_length=6, time_dim_length=14, horizontal_offset=2.6,
            time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[-124.26, -123.26, -122.26, -121.26, -120.26, -119.26, -118.26],
                               [-61.26, -60.26, -59.26, -58.26, -57.26, -56.26, -55.26],
                               [1.74, 2.74, 3.74, 4.74, 5.74, 6.74, 7.74]],

                              [[50.74, 51.74, 52.74, 53.74, 54.74, 55.74, 56.74],
                               [113.74, 114.74, 115.74, 116.74, 117.74, 118.74, 119.74],
                               [176.74, 177.74, 178.74, 179.74, 180.74, 181.74, 182.74]],

                              [[225.74, 226.74, 227.74, 228.74, 229.74, 230.74, 231.74],
                               [288.74, 289.74, 290.74, 291.74, 292.74, 293.74, 294.74],
                               [351.74, 352.74, 353.74, 354.74, 355.74, 356.74, 357.74]],

                              [[400.74, 401.74, 402.74, 403.74, 404.74, 405.74, 406.74],
                               [463.74, 464.74, 465.74, 466.74, 467.74, 468.74, 469.74],
                               [526.74, 527.74, 528.74, 529.74, 530.74, 531.74, 532.74]],

                              [[575.74, 576.74, 577.74, 578.74, 579.74, 580.74, 581.74],
                               [638.74, 639.74, 640.74, 641.74, 642.74, 643.74, 644.74],
                               [701.74, 702.74, 703.74, 704.74, 705.74, 706.74, 707.74]]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_4d_data_and_small_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(alt_dim_length=4,
                                                                 time_dim_length=3, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(alt_dim_length=4, time_dim_length=3,
                                                               horizontal_offset=0.1, altitude_offset=0.1,
                                                               time_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        assert numpy.array_equal(data_cube.data, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(sample_cube.coord('altitude').points, out_cube.coord('altitude').points)

    @istest
    def test_gridded_gridded_li_with_4d_data_and_small_offset(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            alt_dim_length=2, time_dim_length=3, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            alt_dim_length=2, time_dim_length=3, horizontal_offset=0.1, altitude_offset=0.1,
            time_offset=0.1))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[[0.37714286, 1.37714286, 2.37714286],
                                [3.37714286, 4.37714286, 5.37714286]],

                               [[6.37714286, 7.37714286, 8.37714286],
                                [9.37714286, 10.37714286, 11.37714286]],

                               [[12.37714286, 13.37714286, 14.37714286],
                                [15.37714286, 16.37714286, 17.37714286]]],

                              [[[18.37714286, 19.37714286, 20.37714286],
                                [21.37714286, 22.37714286, 23.37714286]],

                               [[24.37714286, 25.37714286, 26.37714286],
                                [27.37714286, 28.37714286, 29.37714286]],

                               [[30.37714286, 31.37714286, 32.37714286],
                                [33.37714286, 34.37714286, 35.37714286]]],

                              [[[36.37714286, 37.37714286, 38.37714286],
                                [39.37714286, 40.37714286, 41.37714286]],

                               [[42.37714286, 43.37714286, 44.37714286],
                                [45.37714286, 46.37714286, 47.37714286]],

                               [[48.37714286, 49.37714286, 50.37714286],
                                [51.37714286, 52.37714286, 53.37714286]]],

                              [[[54.37714286, 55.37714286, 56.37714286],
                                [57.37714286, 58.37714286, 59.37714286]],

                               [[60.37714286, 61.37714286, 62.37714286],
                                [63.37714286, 64.37714286, 65.37714286]],

                               [[66.37714286, 67.37714286, 68.37714286],
                                [69.37714286, 70.37714286, 71.37714286]]],

                              [[[72.37714286, 73.37714286, 74.37714286],
                                [75.37714286, 76.37714286, 77.37714286]],

                               [[78.37714286, 79.37714286, 80.37714286],
                                [81.37714286, 82.37714286, 83.37714286]],

                               [[84.37714286, 85.37714286, 86.37714286],
                                [87.37714286, 88.37714286, 89.37714286]]]])

        assert numpy.allclose(result, out_cube.data)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(sample_cube.coord('altitude').points, out_cube.coord('altitude').points)

    @istest
    def test_gridded_gridded_li_with_sample_grid_4d_data_grid_with_time_with_moderate_offset_and_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(pres_dim_length=3, time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=10, lat_dim_length=6, time_dim_length=14, horizontal_offset=2.6,
            time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[-124.26, -123.26, -122.26, -121.26, -120.26, -119.26, -118.26],
                               [-61.26, -60.26, -59.26, -58.26, -57.26, -56.26, -55.26],
                               [1.74, 2.74, 3.74, 4.74, 5.74, 6.74, 7.74]],

                              [[50.74, 51.74, 52.74, 53.74, 54.74, 55.74, 56.74],
                               [113.74, 114.74, 115.74, 116.74, 117.74, 118.74, 119.74],
                               [176.74, 177.74, 178.74, 179.74, 180.74, 181.74, 182.74]],

                              [[225.74, 226.74, 227.74, 228.74, 229.74, 230.74, 231.74],
                               [288.74, 289.74, 290.74, 291.74, 292.74, 293.74, 294.74],
                               [351.74, 352.74, 353.74, 354.74, 355.74, 356.74, 357.74]],

                              [[400.74, 401.74, 402.74, 403.74, 404.74, 405.74, 406.74],
                               [463.74, 464.74, 465.74, 466.74, 467.74, 468.74, 469.74],
                               [526.74, 527.74, 528.74, 529.74, 530.74, 531.74, 532.74]],

                              [[575.74, 576.74, 577.74, 578.74, 579.74, 580.74, 581.74],
                               [638.74, 639.74, 640.74, 641.74, 642.74, 643.74, 644.74],
                               [701.74, 702.74, 703.74, 704.74, 705.74, 706.74, 707.74]]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)

    @istest
    def test_gridded_gridded_nn_with_data_grid_4d_sample_grid_with_time_with_moderate_offset_and_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=3, lon_dim_length=3, alt_dim_length=2, time_dim_length=2,
            horizontal_offset=2.6, time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        result = numpy.array([[[[1., 3.],
                                [1., 3.],
                                [1., 3.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.]],

                               [[1., 3.],
                                [1., 3.],
                                [1., 3.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.]],

                               [[5., 7.],
                                [5., 7.],
                                [5., 7.],
                                [6., 8.],
                                [6., 8.],
                                [6., 8.],
                                [6., 8.]]],

                              [[[1., 3.],
                                [1., 3.],
                                [1., 3.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.]],

                               [[1., 3.],
                                [1., 3.],
                                [1., 3.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.],
                                [2., 4.]],

                               [[5., 7.],
                                [5., 7.],
                                [5., 7.],
                                [6., 8.],
                                [6., 8.],
                                [6., 8.],
                                [6., 8.]]],

                              [[[13., 15.],
                                [13., 15.],
                                [13., 15.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.]],

                               [[13., 15.],
                                [13., 15.],
                                [13., 15.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.]],

                               [[17., 19.],
                                [17., 19.],
                                [17., 19.],
                                [18., 20.],
                                [18., 20.],
                                [18., 20.],
                                [18., 20.]]],

                              [[[13., 15.],
                                [13., 15.],
                                [13., 15.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.]],

                               [[13., 15.],
                                [13., 15.],
                                [13., 15.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.],
                                [14., 16.]],

                               [[17., 19.],
                                [17., 19.],
                                [17., 19.],
                                [18., 20.],
                                [18., 20.],
                                [18., 20.],
                                [18., 20.]]],

                              [[[25., 27.],
                                [25., 27.],
                                [25., 27.],
                                [26., 28.],
                                [26., 28.],
                                [26., 28.],
                                [26., 28.]],

                               [[25., 27.],
                                [25., 27.],
                                [25., 27.],
                                [26., 28.],
                                [26., 28.],
                                [26., 28.],
                                [26., 28.]],

                               [[29., 31.],
                                [29., 31.],
                                [29., 31.],
                                [30., 32.],
                                [30., 32.],
                                [30., 32.],
                                [30., 32.]]]])

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(data_cube.coord('altitude').points, out_cube.coord('altitude').points)

    @istest
    def test_gridded_gridded_li_with_data_grid_4d_sample_grid_with_time_with_moderate_offset_and_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=3, lon_dim_length=3, alt_dim_length=2, time_dim_length=2,
            horizontal_offset=2.6, time_offset=1.5))

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        result = numpy.array([[[[-5.7, -4.7, -3.7, -2.7, -1.7, -0.7, 0.3],
                                [-3.7, -2.7, -1.7, -0.7, 0.3, 1.3, 2.3]],

                               [[-1.7, -0.7, 0.3, 1.3, 2.3, 3.3, 4.3],
                                [0.3, 1.3, 2.3, 3.3, 4.3, 5.3, 6.3]],

                               [[2.3, 3.3, 4.3, 5.3, 6.3, 7.3, 8.3],
                                [4.3, 5.3, 6.3, 7.3, 8.3, 9.3, 10.3]]],

                              [[[0.3, 1.3, 2.3, 3.3, 4.3, 5.3, 6.3],
                                [2.3, 3.3, 4.3, 5.3, 6.3, 7.3, 8.3]],

                               [[4.3, 5.3, 6.3, 7.3, 8.3, 9.3, 10.3],
                                [6.3, 7.3, 8.3, 9.3, 10.3, 11.3, 12.3]],

                               [[8.3, 9.3, 10.3, 11.3, 12.3, 13.3, 14.3],
                                [10.3, 11.3, 12.3, 13.3, 14.3, 15.3, 16.3]]],

                              [[[6.3, 7.3, 8.3, 9.3, 10.3, 11.3, 12.3],
                                [8.3, 9.3, 10.3, 11.3, 12.3, 13.3, 14.3]],

                               [[10.3, 11.3, 12.3, 13.3, 14.3, 15.3, 16.3],
                                [12.3, 13.3, 14.3, 15.3, 16.3, 17.3, 18.3]],

                               [[14.3, 15.3, 16.3, 17.3, 18.3, 19.3, 20.3],
                                [16.3, 17.3, 18.3, 19.3, 20.3, 21.3, 22.3]]],

                              [[[12.3, 13.3, 14.3, 15.3, 16.3, 17.3, 18.3],
                                [14.3, 15.3, 16.3, 17.3, 18.3, 19.3, 20.3]],

                               [[16.3, 17.3, 18.3, 19.3, 20.3, 21.3, 22.3],
                                [18.3, 19.3, 20.3, 21.3, 22.3, 23.3, 24.3]],

                               [[20.3, 21.3, 22.3, 23.3, 24.3, 25.3, 26.3],
                                [22.3, 23.3, 24.3, 25.3, 26.3, 27.3, 28.3]]],

                              [[[18.3, 19.3, 20.3, 21.3, 22.3, 23.3, 24.3],
                                [20.3, 21.3, 22.3, 23.3, 24.3, 25.3, 26.3]],

                               [[22.3, 23.3, 24.3, 25.3, 26.3, 27.3, 28.3],
                                [24.3, 25.3, 26.3, 27.3, 28.3, 29.3, 30.3]],

                               [[26.3, 27.3, 28.3, 29.3, 30.3, 31.3, 32.3],
                                [28.3, 29.3, 30.3, 31.3, 32.3, 33.3, 34.3]]]]).transpose(0, 1, 3, 2)

        assert numpy.allclose(out_cube.data, result)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube.coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(data_cube.coord('altitude').points, out_cube.coord('altitude').points)

    @istest
    def test_gridded_gridded_nn_with_very_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=0, alt_dim_length=10, time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=0, pres_dim_length=6, time_dim_length=15,
            horizontal_offset=2.6, time_offset=1.5))

        # Result should maintain longitude, pressure and time, and discard latitude and altitude

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        # We will not verify the data here, just that the output has the correct shape
        assert out_cube.data.shape == (7, 3, 6)
        assert not does_coord_exist_in_cube(out_cube, 'latitude')
        assert not does_coord_exist_in_cube(out_cube, 'altitude')
        assert numpy.array_equal(data_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(data_cube.coord('air_pressure').points, out_cube.coord('air_pressure').points)

    @istest
    def test_gridded_gridded_li_with_very_different_grids(self):
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=0, alt_dim_length=10, time_dim_length=7, data_offset=1.0))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=0, pres_dim_length=6, time_dim_length=15,
            horizontal_offset=2.6, time_offset=1.5))

        # Result should maintain time, longitude and pressure (in that order) and discard latitude and altitude

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        # We will not verify the data here, just that the output has the correct shape
        assert out_cube.data.shape == (7, 3, 6)
        assert not does_coord_exist_in_cube(out_cube, 'latitude')
        assert not does_coord_exist_in_cube(out_cube, 'altitude')
        assert numpy.array_equal(data_cube.coord('longitude').points, out_cube.coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube.coord('time').points)
        assert numpy.array_equal(data_cube.coord('air_pressure').points, out_cube.coord('air_pressure').points)

    # Tests for missing values in sample data - nearest neighbour kernel
    @istest
    def test_gridded_gridded_nn_with_different_grids_and_missing_sample_values_in_2d(self):
        mask = [[False, False, False, False, False],
                [False, True, False, True, False],
                [False, False, True, False, False]]
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=5, lon_dim_length=3, dim_order=['lon', 'lat'], mask=mask))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            time_dim_length=4, dim_order=['time', 'lat', 'lon']))

        # Result should have longitude, latitude and time.

        col = GriddedCollocator(missing_data_for_missing_sample=True)

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        # Check that the output has the correct shape and missing values.
        # The order of the dimensions is that of sample with any dimensions not in the sample cube at the end.
        assert out_cube.data.shape == (3, 5, 4)
        assert out_cube.data[1, 1, 0] is numpy.ma.masked
        assert out_cube.data[1, 1, 1] is numpy.ma.masked
        assert out_cube.data[1, 1, 2] is numpy.ma.masked
        assert out_cube.data[1, 1, 3] is numpy.ma.masked
        assert out_cube.data[2, 2, 0] is numpy.ma.masked
        assert out_cube.data[2, 2, 1] is numpy.ma.masked
        assert out_cube.data[2, 2, 2] is numpy.ma.masked
        assert out_cube.data[2, 2, 3] is numpy.ma.masked
        assert out_cube.data[1, 3, 0] is numpy.ma.masked
        assert out_cube.data[1, 3, 1] is numpy.ma.masked
        assert out_cube.data[1, 3, 2] is numpy.ma.masked
        assert out_cube.data[1, 3, 3] is numpy.ma.masked
        c = 0
        for m in out_cube.data.mask.ravel():
            if m:
                c += 1
        assert c == 12

    @istest
    def test_gridded_gridded_nn_with_different_grids_and_missing_sample_values_in_3d(self):
        mask = [[[False, False, False, False, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False]],
                [[False, True, False, True, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False]],
                [[False, False, True, False, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False],
                 [False, False, False, False, False]]]
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=5, lon_dim_length=3, time_dim_length=4,
            dim_order=['lon', 'time', 'lat'], mask=mask))
        data_cube = gridded_data.make_from_cube(make_mock_cube(dim_order=['lat', 'lon']))

        # Result should have latitude and longitude.

        col = GriddedCollocator(missing_data_for_missing_sample=True)

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_nn())[0]

        # Check that the output has the correct shape and missing values.
        # The order of the dimensions is that of sample with any dimensions not in the sample cube at the end.
        assert out_cube.data.shape == (3, 5)
        # No masking since time coordinate has been collapsed.
        c = 0
        if numpy.ma.getmask(out_cube.data) is not numpy.ma.nomask:
            for m in out_cube.data.mask.ravel():
                if m:
                    c += 1
            assert c == 0

    # Tests for missing values in sample data - nearest neighbour kernel
    @istest
    def test_gridded_gridded_li_with_different_grids_and_missing_sample_values_1(self):
        mask = [[False, False, False, False, False],
                [False, True, False, True, False],
                [False, False, False, False, False]]
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=5, lon_dim_length=3, dim_order=['lon', 'lat'], mask=mask))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            time_dim_length=4, dim_order=['lat', 'time', 'lon']))

        # Result should have longitude, latitude and time.

        col = GriddedCollocator(missing_data_for_missing_sample=True)

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert out_cube.data.shape == (3, 5, 4)
        assert out_cube.data[1, 1, 0] is numpy.ma.masked
        assert out_cube.data[1, 1, 1] is numpy.ma.masked
        assert out_cube.data[1, 1, 2] is numpy.ma.masked
        assert out_cube.data[1, 1, 3] is numpy.ma.masked
        assert out_cube.data[1, 3, 0] is numpy.ma.masked
        assert out_cube.data[1, 3, 1] is numpy.ma.masked
        assert out_cube.data[1, 3, 2] is numpy.ma.masked
        assert out_cube.data[1, 3, 3] is numpy.ma.masked
        c = 0
        for m in out_cube.data.mask.ravel():
            if m:
                c += 1
        assert c == 8

    @istest
    def test_gridded_gridded_li_with_different_grids_and_missing_sample_values_2(self):
        mask = [[False, False, False, False, False],
                [False, True, False, True, False],
                [False, False, False, False, False]]
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=5, lon_dim_length=3, dim_order=['lon', 'lat'], mask=mask))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            time_dim_length=4, dim_order=['time', 'lat', 'lon']))

        # Result should have longitude, latitude and time.

        col = GriddedCollocator(missing_data_for_missing_sample=True)

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert out_cube.data.shape == (3, 5, 4)
        assert out_cube.data[1, 1, 0] is numpy.ma.masked
        assert out_cube.data[1, 1, 1] is numpy.ma.masked
        assert out_cube.data[1, 1, 2] is numpy.ma.masked
        assert out_cube.data[1, 1, 3] is numpy.ma.masked
        assert out_cube.data[1, 3, 0] is numpy.ma.masked
        assert out_cube.data[1, 3, 1] is numpy.ma.masked
        assert out_cube.data[1, 3, 2] is numpy.ma.masked
        assert out_cube.data[1, 3, 3] is numpy.ma.masked
        c = 0
        for m in out_cube.data.mask.ravel():
            if m:
                c += 1
        assert c == 8

    @istest
    def test_gridded_gridded_li_with_different_grids_and_missing_sample_values_3(self):
        mask = [[False, False, False],
                [False, False, False],
                [False, False, False],
                [False, False, True],
                [False, False, False]]
        sample_cube = gridded_data.make_from_cube(make_mock_cube(
            lat_dim_length=5, lon_dim_length=3, dim_order=['lat', 'lon'], mask=mask))
        data_cube = gridded_data.make_from_cube(make_mock_cube(
            time_dim_length=4, pres_dim_length=2, dim_order=['lon', 'pres', 'lat', 'time']))

        # Result should have latitude, longitude, pressure and time.

        col = GriddedCollocator(missing_data_for_missing_sample=True)

        out_cube = col.collocate(points=sample_cube, data=data_cube, constraint=None, kernel=gridded_gridded_li())[0]

        assert out_cube.data.shape == (5, 3, 2, 4)
        assert out_cube.data[3, 2, 0, 0] is numpy.ma.masked
        assert out_cube.data[3, 2, 0, 1] is numpy.ma.masked
        assert out_cube.data[3, 2, 0, 2] is numpy.ma.masked
        assert out_cube.data[3, 2, 0, 3] is numpy.ma.masked
        assert out_cube.data[3, 2, 1, 0] is numpy.ma.masked
        assert out_cube.data[3, 2, 1, 1] is numpy.ma.masked
        assert out_cube.data[3, 2, 1, 2] is numpy.ma.masked
        assert out_cube.data[3, 2, 1, 3] is numpy.ma.masked
        c = 0
        for m in out_cube.data.mask.ravel():
            if m:
                c += 1
        assert c == 8

    def test_gridded_gridded_li_for_GriddedDataList(self):
        from cis.data_io.gridded_data import GriddedDataList

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))
        data_cube1 = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=10, lat_dim_length=6, time_dim_length=14, horizontal_offset=2.6,
            time_offset=1.5))
        data_cube2 = gridded_data.make_from_cube(make_mock_cube(
            lon_dim_length=10, lat_dim_length=6, time_dim_length=14, horizontal_offset=2.6,
            time_offset=1.5, data_offset=3))
        data_list = GriddedDataList([data_cube1, data_cube2])

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_list, constraint=None, kernel=gridded_gridded_li())

        result1 = numpy.array([[[-124.26, -123.26, -122.26, -121.26, -120.26, -119.26, -118.26],
                                [-61.26, -60.26, -59.26, -58.26, -57.26, -56.26, -55.26],
                                [1.74, 2.74, 3.74, 4.74, 5.74, 6.74, 7.74]],

                               [[50.74, 51.74, 52.74, 53.74, 54.74, 55.74, 56.74],
                                [113.74, 114.74, 115.74, 116.74, 117.74, 118.74, 119.74],
                                [176.74, 177.74, 178.74, 179.74, 180.74, 181.74, 182.74]],

                               [[225.74, 226.74, 227.74, 228.74, 229.74, 230.74, 231.74],
                                [288.74, 289.74, 290.74, 291.74, 292.74, 293.74, 294.74],
                                [351.74, 352.74, 353.74, 354.74, 355.74, 356.74, 357.74]],

                               [[400.74, 401.74, 402.74, 403.74, 404.74, 405.74, 406.74],
                                [463.74, 464.74, 465.74, 466.74, 467.74, 468.74, 469.74],
                                [526.74, 527.74, 528.74, 529.74, 530.74, 531.74, 532.74]],

                               [[575.74, 576.74, 577.74, 578.74, 579.74, 580.74, 581.74],
                                [638.74, 639.74, 640.74, 641.74, 642.74, 643.74, 644.74],
                                [701.74, 702.74, 703.74, 704.74, 705.74, 706.74, 707.74]]])

        result2 = result1 + 3

        assert numpy.allclose(out_cube[0].data, result1)
        assert numpy.allclose(out_cube[1].data, result2)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube[0].coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube[0].coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube[0].coord('time').points)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube[1].coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube[1].coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube[1].coord('time').points)

    @istest
    def test_gridded_gridded_nn_for_GriddedDataList(self):
        from cis.data_io.gridded_data import GriddedDataList

        sample_cube = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=1.0))

        data_cube1 = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7,
                                                                horizontal_offset=2.6, time_offset=1.5))
        data_cube2 = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, data_offset=3,
                                                                horizontal_offset=2.6, time_offset=1.5))
        data_list = GriddedDataList([data_cube1, data_cube2])

        col = self.collocator

        out_cube = col.collocate(points=sample_cube, data=data_list, constraint=None, kernel=gridded_gridded_nn())

        result = numpy.array([[[1., 1., 1., 2., 3., 4., 5.],
                               [1., 1., 1., 2., 3., 4., 5.],
                               [8., 8., 8., 9., 10., 11., 12.]],

                              [[1., 1., 1., 2., 3., 4., 5.],
                               [1., 1., 1., 2., 3., 4., 5.],
                               [8., 8., 8., 9., 10., 11., 12.]],

                              [[22., 22., 22., 23., 24., 25., 26.],
                               [22., 22., 22., 23., 24., 25., 26.],
                               [29., 29., 29., 30., 31., 32., 33.]],

                              [[43., 43., 43., 44., 45., 46., 47.],
                               [43., 43., 43., 44., 45., 46., 47.],
                               [50., 50., 50., 51., 52., 53., 54.]],

                              [[64., 64., 64., 65., 66., 67., 68.],
                               [64., 64., 64., 65., 66., 67., 68.],
                               [71., 71., 71., 72, 73., 74., 75.]]])

        assert numpy.allclose(out_cube[0].data, result)
        assert numpy.allclose(out_cube[1].data, result + 3)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube[0].coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube[0].coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube[0].coord('time').points)
        assert numpy.array_equal(sample_cube.coord('latitude').points, out_cube[1].coord('latitude').points)
        assert numpy.array_equal(sample_cube.coord('longitude').points, out_cube[1].coord('longitude').points)
        assert numpy.array_equal(sample_cube.coord('time').points, out_cube[1].coord('time').points)

    def test_gridded_gridded_li_when_grids_have_different_dims_order(self):
        # JASCIS-201
        sample = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, dim_order=['lat', 'lon', 'time']))
        data = gridded_data.make_from_cube(make_mock_cube(lat_dim_length=11, lon_dim_length=13,
                                                          time_dim_length=10, dim_order=['time', 'lon', 'lat']))

        col = self.collocator
        out_cube = col.collocate(points=sample, data=data, constraint=None, kernel=gridded_gridded_li())
        assert out_cube[0].shape == sample.shape

    def test_gridded_gridded_nn_when_grids_have_different_dims_order(self):
        # JASCIS-201
        sample = gridded_data.make_from_cube(make_mock_cube(time_dim_length=7, dim_order=['lat', 'lon', 'time']))
        data = gridded_data.make_from_cube(make_mock_cube(lat_dim_length=11, lon_dim_length=13,
                                                          time_dim_length=10, dim_order=['time', 'lon', 'lat']))

        col = self.collocator
        out_cube = col.collocate(points=sample, data=data, constraint=None, kernel=gridded_gridded_nn())
        assert out_cube[0].shape == sample.shape



"""
A heatmap plot (pcolormesh)
"""
from cis.plotting.genericplot import Generic2DPlot
from iris.coords import BOUND_MODE


class Heatmap(Generic2DPlot):

    MODE = BOUND_MODE

    def __init__(self, packed_data, *args, **kwargs):
        # Do this here because if this is ungridded data, we won't be able to complete the super() call
        super(Heatmap, self).__init__(packed_data, *args, **kwargs)

    def __call__(self, ax):
        """
        Plots a heatmap
        """
        self.mappable = ax.pcolormesh(self.x, self.y, self.data, *self.mplargs, **self.mplkwargs)

        super(Heatmap, self).__call__(ax)



import numpy
import unittest
from nose.tools import eq_, raises

from cis.utils import *
from cis.test.utils_for_testing import compare_masked_arrays


class TestUtils(unittest.TestCase):
    def test_can_apply_intersection_mask_to_two_masked_arrays(self):
        import numpy.ma as ma
        import numpy as np

        array1 = ma.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], mask=[1, 1, 1, 0, 0, 0, 0, 0, 0, 0])
        array2 = ma.array([2, 4, 5, 6, 7, 8, 4, 3, 6, 80], mask=[0, 1, 0, 0, 0, 0, 0, 1, 0, 0])
        array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
        assert (np.equal(array1.mask, [1, 1, 1, 0, 0, 0, 0, 1, 0, 0]).all())
        assert (np.equal(array2.mask, [1, 1, 1, 0, 0, 0, 0, 1, 0, 0]).all())

    def test_can_apply_intersection_mask_to_three_masked_arrays(self):
        import numpy.ma as ma

        array1 = ma.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], mask=[1, 1, 1, 0, 0, 0, 0, 0, 0, 0])
        array2 = ma.array([2, 4, 5, 6, 7, 8, 4, 3, 6, 80], mask=[0, 1, 0, 0, 0, 0, 0, 1, 0, 0])
        array3 = ma.array([2, 4, 5, 6, 7, 8, 4, 3, 6, 80], mask=[0, 0, 0, 0, 0, 0, 0, 0, 0, 1])
        array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
        array1, array3 = apply_intersection_mask_to_two_arrays(array1, array3)
        array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)

        assert (ma.equal(array1.mask, [1, 1, 1, 0, 0, 0, 0, 1, 0, 1]).all())
        assert (ma.equal(array2.mask, [1, 1, 1, 0, 0, 0, 0, 1, 0, 1]).all())
        assert (ma.equal(array3.mask, [1, 1, 1, 0, 0, 0, 0, 1, 0, 1]).all())

    def test_can_apply_intersection_mask_to_one_masked_and_one_unmasked_array(self):
        import numpy.ma as ma
        import numpy as np

        array1 = ma.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], mask=[1, 1, 1, 0, 0, 0, 0, 0, 0, 0])
        array2 = np.array([2, 4, 5, 6, 7, 8, 4, 3, 6, 80])
        array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
        assert (np.equal(array1.mask, [1, 1, 1, 0, 0, 0, 0, 0, 0, 0]).all())
        assert (np.equal(array2.mask, [1, 1, 1, 0, 0, 0, 0, 0, 0, 0]).all())

    def test_can_apply_intersection_mask_to_two_unmasked_arrays(self):
        import numpy as np

        array1 = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        array2 = np.array([2, 4, 5, 6, 7, 8, 4, 3, 6, 80])
        array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
        assert (all(array1.mask) is False)
        assert (all(array2.mask) is False)

    def test_can_expand_1d_array_across(self):
        import numpy as np
        from cis.utils import expand_1d_to_2d_array

        a = np.array([1, 2, 3, 4])
        b = expand_1d_to_2d_array(a, 5, axis=0)
        ref = np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
        assert (np.equal(b, ref).all())

    def test_can_expand_1d_array_down(self):
        import numpy as np
        from cis.utils import expand_1d_to_2d_array

        a = np.array([1, 2, 3, 4])
        b = expand_1d_to_2d_array(a, 5, axis=1)
        ref = np.array([[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, 4, 4]])
        assert (np.equal(b, ref).all())

    @raises(ValueError)
    def test_changing_an_element_of_expanded_array_raises_error(self):
        import numpy as np
        from cis.utils import expand_1d_to_2d_array

        a = np.array([1, 2, 3, 4])
        b = expand_1d_to_2d_array(a, 5, axis=0)
        b[1,4] = 42

    def ten_bins_are_created_by_default(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        val_range = {}
        step = None

        bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
        eq_(len(bin_edges), 11)  # 11 edges = 10 bins
        eq_(bin_edges.min(), data.min())
        assert (abs(bin_edges.max() - data.max()) < 1.e-7)

    def bin_width_can_be_specified_where_bin_width_perfectly_divides_range(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        val_range = {}
        step = 0.5

        bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
        eq_(len(bin_edges), 7)
        eq_(bin_edges.min(), data.min())
        eq_(bin_edges.max(), data.max())

    def bin_width_can_be_specified_where_bin_width_does_not_perfectly_divides_range(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        step = 0.7

        bin_edges = calculate_histogram_bin_edges(data, "x", None, None, step)
        eq_(len(bin_edges), 5)
        eq_(bin_edges.min(), data.min())
        assert (bin_edges.max() < data.max())

    def ten_bins_are_created_when_min_is_specified(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        step = None

        bin_edges = calculate_histogram_bin_edges(data, "x", 0.3, None, step)
        eq_(len(bin_edges), 11)  # 11 edges = 10 bins
        eq_(bin_edges.min(), 0.3)
        assert (abs(bin_edges.max() - data.max()) < 1.e-7)  # 1.e-7 is approx 0

    def ten_bins_are_created_when_max_is_specified(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        step = None

        bin_edges = calculate_histogram_bin_edges(data, "x", None, 2.3, step)
        eq_(len(bin_edges), 11)  # 11 edges = 10 bins
        eq_(bin_edges.min(), data.min())
        assert (abs(bin_edges.max() - 2.3) < 1.e-7)  # 1.e-7 is approx 0'''

    def ten_bins_are_created_when_min_and_max_is_specified(self):
        from numpy import array

        data = array([0.0, 1.0, 2.0, 3.0])
        step = None

        bin_edges = calculate_histogram_bin_edges(data, "x", 0.3, 2.3, step)
        eq_(len(bin_edges), 11)  # 11 edges = 10 bins
        assert (abs(bin_edges.min() - 0.3) < 1.e-7)  # 1.e-7 is approx 0
        assert (abs(bin_edges.max() - 2.3) < 1.e-7)  # 1.e-7 is approx 0

    def test_split_into_float_and_units(self):
        eq_(split_into_float_and_units('10km')['value'], 10)
        eq_(split_into_float_and_units('10km')['units'], 'km')

    def test_split_into_float_and_units_with_spaces(self):
        eq_(split_into_float_and_units('10 km')['value'], 10)
        eq_(split_into_float_and_units('10 km')['units'], 'km')

    def test_split_into_float_and_units_with_full_float(self):
        eq_(split_into_float_and_units('12.3e4uM')['value'], 12.3e4)
        eq_(split_into_float_and_units('12.3e4uM')['units'], 'uM')

    @raises(InvalidCommandLineOptionError)
    def test_split_into_float_and_units_with_extra_numbers(self):
        split_into_float_and_units('10km10')

    @raises(InvalidCommandLineOptionError)
    def test_split_into_float_and_units_with_no_numbers(self):
        split_into_float_and_units('km')

    def test_split_into_float_and_units_with_no_units(self):
        eq_(split_into_float_and_units('10')['value'], 10)
        eq_(split_into_float_and_units('10')['units'], None)

    @raises(InvalidCommandLineOptionError)
    def test_split_into_float_and_units_with_extra_units(self):
        eq_(split_into_float_and_units('km10m'), 10)

    def test_parse_distance_with_units_of_km_to_float_km(self):
        eq_(parse_distance_with_units_to_float_km('10km'), 10)

    def test_parse_distance_with_units_of_m_to_float_km(self):
        eq_(parse_distance_with_units_to_float_km('10000m'), 10)

    def test_parse_distance_without_units_to_float_km(self):
        eq_(parse_distance_with_units_to_float_km('10'), 10)

    @raises(InvalidCommandLineOptionError)
    def test_parse_distance_with_invalid_units(self):
        eq_(parse_distance_with_units_to_float_km('10Gb'), 10)

    def test_parse_distance_with_units_of_m_to_float_m(self):
        eq_(parse_distance_with_units_to_float_m('10m'), 10)

    def test_parse_distance_with_units_of_km_to_float_m(self):
        eq_(parse_distance_with_units_to_float_m('10km'), 10000)

    def test_parse_distance_without_units_to_float_m(self):
        eq_(parse_distance_with_units_to_float_m('10'), 10)

    def test_array_equal_including_nan(self):
        array1 = numpy.array([[1, 2], [3, 4]])
        array2 = numpy.array([[1, 2], [3, 4.1]])
        array3 = numpy.array([[1, 2], [3, float('nan')]])
        assert array_equal_including_nan(array1, array1)
        assert not array_equal_including_nan(array1, array2)
        assert not array_equal_including_nan(array1, array3)
        assert array_equal_including_nan(array3, array3)

    # Tests for apply_mask_to_numpy_array

    def test_apply_mask_to_numpy_array_with_unmasked_array(self):
        # Input array not a masked array to which is applied a mask containing 'True's
        in_array = numpy.array([1, 2, 3, 4])
        mask = numpy.array([False, False, True, False])
        out_array = apply_mask_to_numpy_array(in_array, mask)
        assert numpy.array_equal(out_array.mask, mask)

    def test_apply_mask_to_numpy_array_with_masked_array(self):
        # Input array has masked points to which is applied a mask containing 'True's
        in_array = numpy.ma.array([1, 2, 3, 4], mask=numpy.array([True, False, False, False]))
        mask = numpy.array([False, False, True, False])
        out_array = apply_mask_to_numpy_array(in_array, mask)
        assert numpy.array_equal(out_array.mask, numpy.array([True, False, True, False]))

    def test_apply_mask_to_numpy_array_with_masked_array_with_nomask(self):
        # Input array is a masked array but with mask 'nomask'. This is masked by a mask with no 'True's.
        # The output array should not have had a mask created.
        in_array = numpy.ma.array([1, 2, 3, 4], mask=numpy.ma.nomask)
        mask = numpy.array([False, False, False, False])
        out_array = apply_mask_to_numpy_array(in_array, mask)
        assert numpy.ma.getmask(out_array) is numpy.ma.nomask

    def test_apply_mask_to_numpy_array_with_masked_array_but_all_unmasked(self):
        # Input array is a masked array but no elements are masked. The mask contains no 'True's.
        # This is masked by a mask with no 'True's. The output array should not have had a mask created.
        in_array = numpy.ma.array([1, 2, 3, 4], mask=numpy.array([False, False, False, False]))
        mask = numpy.array([False, False, False, False])
        out_array = apply_mask_to_numpy_array(in_array, mask)
        assert numpy.array_equal(out_array.mask, mask)

    def test_apply_mask_to_numpy_array_with_masked_array_with_array_unmasked(self):
        # Input array is a masked array but with mask 'nomask'. This is masked by a mask with 'True's.
        # The output array should not have had a mask created.
        in_array = numpy.ma.array([1, 2, 3, 4], mask=numpy.ma.nomask)
        mask = numpy.array([False, True, False, True])
        out_array = apply_mask_to_numpy_array(in_array, mask)
        assert numpy.array_equal(out_array.mask, mask)

    def test_GIVEN_masked_array_WHEN_fix_longitude_THEN_longitudes_fixed_but_masked_values_untouched(self):
        lons = numpy.ma.masked_less([0, -639, 90, 180, -639, 270, 360, -639], -360)
        new_lons = fix_longitude_range(lons, -180)
        expected_new_lons = numpy.ma.masked_invalid([0, float('Nan'), 90, -180, float('Nan'), -90, 0, float('Nan')])
        compare_masked_arrays(new_lons, expected_new_lons)

    def test_GIVEN_unmasked_array_WHEN_fix_longitude_THEN_longitudes_fixed(self):
        lons = numpy.array([0, 90, 180, 270, 360])
        new_lons = fix_longitude_range(lons, -180)
        assert numpy.array_equal(new_lons, [0, 90, -180, -90, 0])

    def test_GIVEN_lons_write_only_WHEN_fix_longitude_THEN_longitudes_fixed(self):
        # Previously there was an issue where if the longitudes array being read in
        # was read only then this method failed.
        lons = numpy.array([0, 90, 180, 270, 360])
        lons.setflags(write=False)
        new_lons = fix_longitude_range(lons, -180)
        assert numpy.array_equal(new_lons, [0, 90, -180, -90, 0])

    def test_GIVEN_list_of_arrays_to_concatenate_WHEN_all_are_masked_THEN_returns_masked_array_with_correct_mask(self):
        arrays = []
        for i in range(0, 3):
            arrays.append(numpy.ma.array([0, 90, 180], mask=[False, True, False]))
        conc = concatenate(arrays)
        assert numpy.ma.count_masked(conc) == 3

    def test_GIVEN_list_of_arrays_to_concatenate_WHEN_none_are_masked_THEN_returns_numpy_array(self):
        arrays = []
        for i in range(0, 3):
            arrays.append(numpy.array([0, 90, 180]))
        conc = concatenate(arrays)
        assert isinstance(conc, numpy.ndarray)

    def test_GIVEN_list_of_arrays_to_concat_WHEN_some_masked_but_not_first_THEN_returns_masked_array_with_correct_mask(
            self):
        arrays = [numpy.array([0, 90, 180])]
        for i in range(1, 3):
            arrays.append(numpy.ma.array([0, 90, 180], mask=[False, True, False]))
        conc = concatenate(arrays)
        assert numpy.ma.count_masked(conc) == 2

    def test_GIVEN_list_of_arrays_to_concatenate_WHEN_some_are_masked_THEN_returns_masked_array(self):
        arrays = [numpy.ma.array([0, 90, 180], mask=[False, True, False])]
        for i in range(0, 3):
            arrays.append(numpy.array([0, 90, 180]))
        conc = concatenate(arrays)
        assert numpy.ma.count_masked(conc) == 1


class TestFindLongitudeWrapStart(unittest.TestCase):

    def test_GIVEN_data_is_minus_180_to_180_THEN_returns_minus_180(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.)

        eq_(find_longitude_wrap_start(data), -180)

    def test_GIVEN_data_is_minus_0_to_360_THEN_returns_0(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=5, lon_max=345.)

        eq_(find_longitude_wrap_start(data), 0)



#!/usr/bin/env python
# (C) British Crown Copyright 2010 - 2014, Met Office
#
# This file was heavily influenced by a similar file in the iris package.
"""
Provides "diff-like" comparison of images.

Currently relies on matplotlib for image processing so limited to PNG format.

"""

from __future__ import (absolute_import, division, print_function)

import os.path
import shutil

import matplotlib.pyplot as plt
import matplotlib.image as mimg
import matplotlib.widgets as mwidget


def diff_viewer(expected_fname, result_fname, diff_fname):
    plt.figure(figsize=(16, 16))
    plt.suptitle(os.path.basename(expected_fname))
    ax = plt.subplot(221)
    ax.imshow(mimg.imread(expected_fname))
    ax = plt.subplot(222, sharex=ax, sharey=ax)
    ax.imshow(mimg.imread(result_fname))
    ax = plt.subplot(223, sharex=ax, sharey=ax)
    ax.imshow(mimg.imread(diff_fname))

    def accept(event):
        # removes the expected result, and move the most recent result in
        print('ACCEPTED NEW FILE: %s' % (os.path.basename(expected_fname), ))
        os.remove(expected_fname)
        shutil.copy2(result_fname, expected_fname)
        os.remove(diff_fname)
        plt.close()

    def reject(event):
        print('REJECTED: %s' % (os.path.basename(expected_fname), ))
        plt.close()

    ax_accept = plt.axes([0.6, 0.35, 0.1, 0.075])
    ax_reject = plt.axes([0.71, 0.35, 0.1, 0.075])
    bnext = mwidget.Button(ax_accept, 'Accept change')
    bnext.on_clicked(accept)
    bprev = mwidget.Button(ax_reject, 'Reject')
    bprev.on_clicked(reject)

    plt.show()


def step_over_diffs():
    import cis.test.plot_tests
    image_dir = os.path.join(os.path.dirname(cis.test.plot_tests.__file__),
                             'reference', 'visual_tests')
    diff_dir = os.path.join(os.path.dirname(cis.test.plot_tests.__file__),
                            'result_image_comparison')

    for expected_fname in sorted(os.listdir(image_dir)):
        result_path = os.path.join(diff_dir, expected_fname)
        diff_path = result_path[:-4] + '-failed-diff.png'

        # if the test failed, there will be a diff file
        if os.path.exists(diff_path):
            expected_path = os.path.join(image_dir, expected_fname)
            diff_viewer(expected_path, result_path, diff_path)


if __name__ == '__main__':
    step_over_diffs()



from abc import ABCMeta, abstractmethod
import logging
import traceback
from cis.exceptions import FileFormatError
from cis.data_io.hdf import get_hdf4_file_variables
from cis.data_io.netcdf import get_netcdf_file_variables, remove_variables_with_non_spatiotemporal_dimensions
from cis import __version__
import six


@six.add_metaclass(ABCMeta)
class AProduct(object):
    """
    Abstract class for the various possible data products. This just defines the interface which
    the subclasses must implement.
    """

    # If a filename matches two data products' file signatures, the data product with the higher priority will be used
    priority = 10

    # Contains a list of valid spatiotemporal variable names
    valid_dimensions = None

    @abstractmethod
    def create_data_object(self, filenames, variable):
        """
        Create and return an :class:`.CommonData` object for a given variable from one or more files.

        :param list filenames: List of filenames of files to read
        :param str variable: Variable to read from the files
        :return: An :class:`.CommonData` object representing the specified variable

        :raise FileIOError: Unable to read a file
        :raise InvalidVariableError: Variable not present in file
        """

    @abstractmethod
    def create_coords(self, filenames):
        """
        Reads the coordinates from one or more files. Note that this method may have to make certain assumptions about
        the file in order to return a single coordinate set. The user should be warned through the logger if this is the
        case.

        :param list filenames: List of filenames to read coordinates from
        :return: :class:`.CommonData` object
        """

    @abstractmethod
    def get_file_signature(self):
        """
        This method should return a list of regular expressions, which CIS uses to decide which data
        product to use for a given file. If more than one regular expression is provided in the list then the file can
        match `any` of the expressions. The first product with a signature that matches the filename will be used.
        The order in which the products are searched is determined by the priority property, highest value first;
        internal products generally have a priority of 10.

        For example, this would match all files with a name containing the string 'CODE' and with the 'nc' extension.::

            return [r'.*CODE*.nc']

        .. note::

            If the signature has matched the framework will call :meth:`.AProduct.get_file_type_error`, this gives the
            product a chance to open the file and check the contents.

        :return: A list of regex to match the product's file naming convention.
        :rtype: list
        """

    def get_variable_names(self, filenames, data_type=None):
        """
        Get a list of available variable names from the filenames list passed in. This general implementation can be
        overridden in specific products to include/exclude variables which may or may not be relevant.
        The data_type parameter can be used to specify extra information.

        :param list filenames: List of string filenames of files to be read from
        :param str data_type: 'SD' or 'VD' to specify only return SD or VD variables from HDF files. This may take on
         other values in specific product implementations.
        :return: A set of variable names as strings
        :rtype: str
        """
        variables = []
        for filename in filenames:
            file_variables = None
            if filename.endswith(".nc"):
                file_variables = get_netcdf_file_variables(filename)
                remove_variables_with_non_spatiotemporal_dimensions(file_variables, self.valid_dimensions)
            elif filename.endswith(".hdf"):
                if data_type is None:
                    data_type = "SD"
                sd_vars, vd_vars = get_hdf4_file_variables(filename, data_type)
                file_variables = list((sd_vars or {}).keys()) + list((vd_vars or {}).keys())
            variables.extend(file_variables)
        return set(variables)

    def get_file_format(self, filename):
        """
        Returns a file format hierarchy separated by slashes, of the form
        ``TopLevelFormat/SubFormat/SubFormat/Version``.
        E.g. ``NetCDF/GASSP/1.0``, ``ASCII/ASCIIHyperpoint`` or ``HDF4/CloudSat``. This is mainly used within the
        ceda_di indexing tool. If not set it will default to the products name.

        A filename of an example file can be provided to enable the determination of, for example, a dataset version
        number.

        :param str filename: Filename of file to be inspected
        :return: File format, of the form ``[parent/]format/specific instance/version``, or the class name
        :rtype: str
        :raises: FileFormatError if there is an error
        """
        return self.__class__.__name__

    def get_file_type_error(self, filename):
        """
        Check a single file to see if it is of the correct type, and if not return
        a list of errors. If the return is None then there are no errors and
        this is the correct data product to use for this file.

        This method gives a mechanism for a data product to identify itself as the
        correct product when a specific enough file signature cannot be provided. For
        example GASSP is a type of NetCDF file and so filenames end with .nc but
        so do other NetCDF files, so the data product opens the file and looks
        for the GASSP version attribute, and if it doesn't find it returns an
        error.

        :param str filename: The filename for the file
        :return: List of errors, or None
        :rtype: list or None
        """
        return None


def __get_class(filename, product=None):
    """
    Identify the subclass of :class:`.AProduct` to a given product name if specified.
    If the product name is not specified, the routine uses the signature (regex)
    given by :meth:`get_file_signature` to infer the product class from the filename.

    Note, only the first filename of the list is use here.

    :param filename: A single filename
    :param product: name of the product
    :return: a subclass of :class:`.AProduct`
    """
    import re
    import os
    import cis.plugin as plugin
    from cis.exceptions import ClassNotFoundError

    # Ensure the filename doesn't include the path
    basename = os.path.basename(filename)

    product_classes = plugin.find_plugin_classes(AProduct, 'cis.data_io.products')
    product_classes = sorted(product_classes, key=lambda cls: cls.priority, reverse=True)

    for cls in product_classes:

        if product is None:
            # search for a pattern that matches file signature
            class_instance = cls()
            patterns = class_instance.get_file_signature()
            for pattern in patterns:
                # Match the pattern - re.I allows for case insensitive matches.
                # Appending '$' to the pattern ensures we match the whole string
                if re.match(pattern+'$', basename, re.I) is not None:
                    logging.debug("Found product class " + cls.__name__ + " matching regex pattern " + pattern)
                    errors = class_instance.get_file_type_error(filename)
                    if errors is None:
                        return cls
                    else:
                        logging.info("Product class {} is not right because {}".format(cls.__name__, errors))
        else:
            # product specified directly
            if product == cls.__name__:
                logging.debug("Selected product class " + cls.__name__)
                return cls
    error_message = "Product cannot be found for given file.\nSupported products and signatures are:\n"
    for cls in product_classes:
        error_message += cls().__class__.__name__ + ": " + str(cls().get_file_signature()) + "\n"
    raise ClassNotFoundError(error_message)


def get_data(filenames, variable, product=None):
    """
    Top level routine for calling the correct product's :meth:`create_data_object` routine.

    :param list filenames: A list of filenames to read data from
    :param str variable: The variable to create the :class:`.CommonData` object from
    :param str product: The product to read data with - this should be a string which matches the name of one of the
     subclasses of :class:`.AProduct`. If none is supplied it is guessed from the filename signature.
    :return: A :class:`.CommonData` variable
    """
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving data using product " + product_cls.__name__ + "...")
    try:
        data = product_cls().create_data_object(filenames, variable)
        return data
    except Exception as e:
        logging.debug("Error in product plugin %s:\n%s" % (product_cls.__name__, traceback.format_exc()))
        raise ProductPluginException("An error occurred retrieving data using the product %s. Check that this "
                                     "is the correct product plugin for your chosen data. Exception was %s: %s."
                                     % (product_cls.__name__, type(e).__name__, e.args[0]), e)


def get_coordinates(filenames, product=None):
    """
    Top level routine for calling the correct product's :meth:`create_coords` routine.

    :param list filenames: A list of filenames to read data from
    :param str product: The product to read data with - this should be a string which matches the name of one of the
     subclasses of AProduct
    :return: A :class:`CoordList` object
    """
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving coordinates using product " + product_cls.__name__)
    try:
        coord = product_cls().create_coords(filenames)
        return coord
    except Exception as e:
        logging.debug("Error in product plugin %s:\n%s" % (product_cls.__name__, traceback.format_exc()))
        raise ProductPluginException("An error occurred retrieving coordinates using the product %s. Check that this "
                                     "is the correct product plugin for your chosen data. Exception was %s: %s."
                                     % (product_cls.__name__, type(e).__name__, e.args[0]), e)


def get_variables(filenames, product=None, data_type=None):
    """
    Top level routine for calling the correct product's :meth:`get_variable_names` routine.

    :param list filenames: A list of filenames to read the variables from
    :param str product: The product to read data with - this should be a string which matches the name of one of the
     subclasses of AProduct
    :return: A set of variable names as strings
    """
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving variables using product " + product_cls.__name__ + "...")
    try:
        variables = product_cls().get_variable_names(filenames, data_type)
        return variables
    except Exception as e:
        logging.debug("Error in product plugin %s:\n%s" % (product_cls.__name__, traceback.format_exc()))
        raise ProductPluginException("An error occurred retrieving variables using the product %s. Check that this "
                                     "is the correct product plugin for your chosen data. Exception was %s: %s."
                                     % (product_cls.__name__, type(e).__name__, e.args[0]), e)


def get_file_format(filenames, product=None):
    """
    Returns the files format of throws FileFormatError if there is an error in the format

    :param list filenames: the filenames to read
    :param str product: the product to use, if not specified search
    :return: File format
    :raises ClassNotFoundError: if there is no reader for this class
    """
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving file format using product " + product_cls.__name__ + "...")
    try:
        file_format = product_cls().get_file_format(filenames[0])
    except Exception as e:
        logging.debug("Error in product plugin %s:\n%s" % (product_cls.__name__, traceback.format_exc()))
        raise ProductPluginException(
            "An error occurred retrieving the file format using the product %s. Check that this "
            "is the correct product plugin for your chosen data. Exception was %s: %s."
            % (product_cls.__name__, type(e).__name__, e.args[0]), e)

    try:
        product_cls().create_coords(filenames)
    except Exception as ex:
        raise FileFormatError(error_list=['Could not read coordinates from the file', ex.args[0]])
    return file_format


def get_product_full_name(filenames, product=None):
    """
    Get the full name of the product which would read this file

    :param list filenames: list of filenames to read
    :param str product: specified product to use
    """

    product_cls = __get_class(filenames[0], product)
    logging.info("Retrieving product fill name using product " + product_cls.__name__ + "...")

    return "CIS/{product}/{version}".format(product=product_cls.__name__, version=__version__)


class ProductPluginException(Exception):
    """
    Represents an error which has occurred inside of a Product plugin
    """

    original_exception = None

    def __init__(self, message, original_exception):
        super(ProductPluginException, self).__init__(message)
        self.original_exception = original_exception
